// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.provider.azure.pubsub;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.springframework.http.HttpStatus;

import javax.servlet.ReadListener;
import javax.servlet.ServletInputStream;
import javax.servlet.http.HttpServletRequest;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.StringReader;
import java.util.Map;

import static org.junit.Assert.fail;
import static org.mockito.Mockito.when;

@RunWith(MockitoJUnitRunner.class)
public class EventGridRequestBodyExtractorTest {

    private EventGridRequestBodyExtractor sut;

    @Mock
    private HttpServletRequest httpServletRequest;

    @Mock
    private JaxRsDpsLog log;

    @Test
    public void should_returnTrue_isHandshakeRequest() throws IOException {
        // Set up
        String validHandshakeRequestRoot =
                "[{\"id\":\"2d1781af-3a4c-4d7c-bd0c-e34b19da4e66\",\"topic\":\"/subscriptions\",\"subject\":\"\",\"data\":{\"validationCode\":\"sample\",\"validationUrl\":\"https://rp-eastus2.eventgrid.azure.net:553/eventsubscriptions/estest/validate?id=sample&t=2018-04-26T20:30:54.4538837Z&apiVersion=2018-05-01-preview&token=1A1A1A1A\"},\"eventType\":\"Microsoft.EventGrid.SubscriptionValidationEvent\",\"eventTime\":\"2018-01-25T22:12:19.4556811Z\",\"metadataVersion\":\"1\",\"dataVersion\":\"1\"}]";
        setServletRequest(validHandshakeRequestRoot);

        // Act
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);
        boolean response = this.sut.isHandshakeRequest();

        // Assert
        Assert.assertTrue(response);
    }

    @Test
    public void shouldThrow_whenRequestTypeIsNotright_isHandshakeRequest() throws IOException {
        //SetUp
        String invalidHandshakeRequestRoot =
                "[{\"id\":\"2d1781af-3a4c-4d7c-bd0c-e34b19da4e66\",\"topic\":\"/subscriptions\",\"subject\":\"\",\"data\":{\"validationCode\":\"sample\",\"validationUrl\":\"sample\"},\"eventType\":\"invalid\",\"eventTime\":\"sample\",\"metadataVersion\":\"1\",\"dataVersion\":\"1\"}]";
        setServletRequest(invalidHandshakeRequestRoot);

        try {
            // Act
            sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

            // Assert
            fail("Should Throw Exception");
        } catch (AppException appException) {
            Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
            Assert.assertEquals("Unable to parse request payload.", appException.getError().getMessage());
        } catch (Exception exception) {
            fail("Should Throw AppException");
        }
    }

    @Test
    public void should_throwWhenAttributesAreMissing_extractDataFromRequestBody() throws IOException {
        String[] requestRootWithoutAttributes = {
                "[{\"id\":\"sampleid\",\"subject\":\"RecordsChanged\",\"data\":{\"account-id\":\"opendes\",\"correlation-id\":\"sample\",\"data-partition-id\":\"opendes\"},\"eventType\":\"RecordsChanged\",\"dataVersion\":\"1.0\",\"metadataVersion\":\"1\",\"eventTime\":\"sample\",\"topic\":\"fhtfc\"}]",
                "[{\"id\":\"sampleid\",\"subject\":\"RecordsChanged\",\"data\":{\"data\":[{\"id\":\"opendes:doc:guid\",\"kind\":\"kimd\",\"op\":\"create\"},{\"id\":\"opendes:doc:\",\"kind\":\"kimd\",\"op\":\"create\"}],\"account-id\":\"opendes\",\"data-partition-id\":\"opendes\"},\"eventType\":\"RecordsChanged\",\"dataVersion\":\"1.0\",\"metadataVersion\":\"1\",\"eventTime\":\"sample\",\"topic\":\"fhtfc\"}]",
                "[{\"id\":\"sampleid\",\"subject\":\"RecordsChanged\",\"data\":{\"data\":[{\"id\":\"opendes:doc:guid\",\"kind\":\"kimd\",\"op\":\"create\"},{\"id\":\"opendes:doc:\",\"kind\":\"kimd\",\"op\":\"create\"}],\"account-id\":\"opendes\",\"correlation-id\":\"sample\"},\"eventType\":\"RecordsChanged\",\"dataVersion\":\"1.0\",\"metadataVersion\":\"1\",\"eventTime\":\"sample\",\"topic\":\"fhtfc\"}]",
                "[{\"id\":\"sampleid\",\"subject\":\"RecordsChanged\",\"eventType\":\"RecordsChanged\",\"dataVersion\":\"1.0\",\"metadataVersion\":\"1\",\"eventTime\":\"sample\",\"topic\":\"fhtfc\"}]"
        };
        for (String request : requestRootWithoutAttributes) {
            setServletRequest(request);
            try {
                // Act
                sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

                // Asset
                fail("Should Throw Exception");
            } catch (AppException appException) {
                Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
                Assert.assertEquals("Unable to parse request payload.", appException.getError().getMessage());
            } catch (Exception exception) {
                fail("Should Throw AppException");
            }
        }
    }

    @Test
    public void should_returnValidResponse_getHandshakeResponse() throws IOException {
        // Set up
        String validHandshakeRequestRoot =
                "    [{\n" +
                        "        \"id\": \"testId\",\n" +
                        "        \"topic\": \"testTopic\",\n" +
                        "        \"subject\": \"\",\n" +
                        "        \"data\": {\n" +
                        "            \"validationCode\": \"testValidationCode\",\n" +
                        "            \"validationUrl\": \"testURL\"\n" +
                        "        },\n" +
                        "        \"eventType\": \"Microsoft.EventGrid.SubscriptionValidationEvent\",\n" +
                        "        \"eventTime\": \"sampletime\",\n" +
                        "        \"metadataVersion\": \"1\",\n" +
                        "        \"dataVersion\": \"2\"\n" +
                        "    }]";
        String expectedResponse = "testValidationCode";
        setServletRequest(validHandshakeRequestRoot);

        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        String observedResponse = this.sut.getValidationCodeForHandshake();

        // Assert
        Assert.assertEquals(observedResponse, expectedResponse);
    }

    @Test
    public void should_throwWhenNotHandshakeRequest_getHandshakeResponse() throws IOException {
        // Set up
        String validHandshakeRequestRoot =
                "[{\"id\":\"sampleid\",\"subject\":\"RecordsChanged\",\"data\":{\"data\":[{\"id\":\"opendes:doc:guid\",\"kind\":\"kimd\",\"op\":\"create\"},{\"id\":\"opendes:doc:\",\"kind\":\"kimd\",\"op\":\"create\"}],\"account-id\":\"opendes\",\"correlation-id\":\"sample\",\"data-partition-id\":\"opendes\"},\"eventType\":\"RecordsChanged\",\"dataVersion\":\"1.0\",\"metadataVersion\":\"1\",\"eventTime\":\"sample\",\"topic\":\"fhtfc\"}]";
        String expectedResponse = null;
        setServletRequest(validHandshakeRequestRoot);

        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        String observedResponse = this.sut.getValidationCodeForHandshake();

        // Assert
        Assert.assertNull(observedResponse);
    }

    @Test
    public void should_throwWhenHandshakeRequest_extractDataFromRequestBody() throws IOException {
        String inValidRequestRoot = "    [{\n" +
                "        \"id\": \"testId\",\n" +
                "        \"topic\": \"testTopic\",\n" +
                "        \"subject\": \"\",\n" +
                "        \"data\": {\n" +
                "            \"validationCode\": \"testValidationCode\",\n" +
                "            \"validationUrl\": \"testURL\"\n" +
                "        },\n" +
                "        \"eventType\": \"Microsoft.EventGrid.SubscriptionValidationEvent\",\n" +
                "        \"eventTime\": \"sampletime\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"dataVersion\": \"2\"\n" +
                "    }]";
        setServletRequest(inValidRequestRoot);
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        Map<String, String> observedAttributes = this.sut.extractAttributesFromRequestBody();

        Assert.assertNull(observedAttributes);
    }

    @Test
    public void should_returnNotificationId_extractNotificationIdFromRequestBody() throws IOException {
        // Set Up
        String validRequestRoot = "[{\"id\":\"sampleid\",\"subject\":\"RecordsChanged\",\"data\":{\"data\":[{\"id\":\"opendes:doc:guid\",\"kind\":\"kimd\",\"op\":\"create\"},{\"id\":\"opendes:doc\",\"kind\":\"kimd\",\"op\":\"create\"}],\"account-id\":\"opendes\",\"correlation-id\":\"sample\",\"data-partition-id\":\"opendes\"},\"eventType\":\"RecordsChanged\",\"dataVersion\":\"1.0\",\"metadataVersion\":\"1\",\"eventTime\":\"sample\",\"topic\":\"fhtfc\"}]";
        setServletRequest(validRequestRoot);

        when(this.httpServletRequest.getHeader("Aeg-Subscription-Name")).thenReturn("NotificationId");
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        String observed = sut.extractNotificationIdFromRequestBody();

        // Assert
        Assert.assertEquals("notificationid", observed);

        // Set Up
        when(this.httpServletRequest.getHeader("Aeg-Subscription-Name")).thenReturn(null);

        try {
            // Act
            this.sut.extractNotificationIdFromRequestBody();

            // Asset
            fail("Should Throw Exception");
        } catch (AppException appException) {
            Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
            Assert.assertEquals("Subscription ID not found", appException.getError().getMessage());
        } catch (Exception exception) {
            fail("Should Throw AppException");
        }
    }

    private void setServletRequest(String validHandshakeRequestRoot) throws IOException {
        BufferedReader reader = new BufferedReader(new StringReader(validHandshakeRequestRoot));

        ByteArrayInputStream byteArrayInputStream = new ByteArrayInputStream(validHandshakeRequestRoot.getBytes());
        when(httpServletRequest.getMethod()).thenReturn("POST");
        when(httpServletRequest.getInputStream()).thenReturn(new ServletInputStream() {
            @Override
            public boolean isFinished() {
                return false;
            }

            @Override
            public boolean isReady() {
                return false;
            }

            @Override
            public void setReadListener(ReadListener readListener) {

            }

            @Override
            public int read() throws IOException {
                return byteArrayInputStream.read();
            }
        });
    }
}
