/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriver;
import org.opengroup.osdu.core.gcp.oqm.model.*;
import org.opengroup.osdu.notification.provider.gcp.config.OqmConfigurationProperties;
import org.opengroup.osdu.notification.provider.gcp.pubsub.di.OqmNotificationHandler;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.springframework.beans.factory.config.BeanDefinition.SCOPE_SINGLETON;

/**
 * Runs once on the service start.
 * 1. Fetches oqm for tenants' message brokers' pull subscriptions in interested topics. Creates the service's subscribers in every found subscription.
 * 2. Checks for the "subscriber control topic" on every tenant's message broker and creates if it is absent.
 * - This topic is a "control channel" between Register and Notification services.
 * - Register sends events on new pull Subscriptions being created, Notification listens for events and creates corresponding Subscribers.
 */
@Slf4j
@Component
@Scope(SCOPE_SINGLETON)
@ConditionalOnProperty(name = "oqmDriver")
@RequiredArgsConstructor
public class OqmSubscriberManager {

    private final String ACKNOWLEDGE = "message acknowledged by client";
    private final String NOT_ACKNOWLEDGE = "message not acknowledged by client";

    //TODO should be externalized to application.properties
    private static final List<OqmTopic> INTERESTED_TOPICS =
            Stream.of("records-changed", "schema-changed", "status-changed", "legaltags_changed")
                    .map(topicName -> OqmTopic.builder().name(topicName).build()).collect(Collectors.toList());

    private static final String INTERESTED_SUBSCRIPTIONS_PREFIX = "de-";

    private static final OqmSubscriptionQuery INTERESTED_SUBSCRIPTIONS_QUERY = OqmSubscriptionQuery.builder()
            .forAnyOfTopics(INTERESTED_TOPICS).namePrefix(INTERESTED_SUBSCRIPTIONS_PREFIX)
            .subscriberable(true).build();

    private final ITenantFactory tenantInfoFactory;
    private final OqmDriver driver;
    private final OqmNotificationHandler notificationHandler;
    private final OqmConfigurationProperties properties;

    private final Long constructDate = System.currentTimeMillis();
    private final String subscriberControlTopicSubscriptionName = "notification-service-" + constructDate;

    @PostConstruct
    void postConstruct() {
        log.debug("OqmSubscriberManager bean constructed. Provisioning STARTED.");

        provisionSubscribersOnAllInterestedTopicsSubscriptionsForAllTenantsBrokers();
        provisionControlTopicsWithSubscriptionsAndSubscribersForAllTenantsBrokers();

        log.debug("OqmSubscriberManager bean constructed. Provisioning COMPLETED.");
    }

    @PreDestroy
    void onPreDestroy() {
        log.debug("OqmSubscriberManager bean on pre-destroy: STARTED.");
        unprovisionControlTopicsSubscriptionsFromAllTenantsBrokers();
        log.debug("OqmSubscriberManager bean on pre-destroy: COMPLETED.");
    }

    void unprovisionControlTopicsSubscriptionsFromAllTenantsBrokers() {
        for (TenantInfo tenant : tenantInfoFactory.listTenantInfo()) {
            String tenantId = tenant.getDataPartitionId();
            log.debug("* OqmSubscriberManager on pre-destroy for tenant {}:", tenantId);
            OqmSubscription subscriberControlTopicSubscriptionForTenant = driver.getSubscription(subscriberControlTopicSubscriptionName, getDestination(tenant)).orElse(null);
            if (subscriberControlTopicSubscriptionForTenant != null) {
                log.debug("* * OqmSubscriberManager on delete '{}' subscriber control topic subscription with name '{}' at tenant's '{}' message broker: Subscription  DELETED.",
                        properties.getRegisterSubscriberControlTopicName(), subscriberControlTopicSubscriptionName, tenantId);
                driver.deleteSubscription(subscriberControlTopicSubscriptionName, getDestination(tenant));
            }
        }
    }

    void provisionSubscribersOnAllInterestedTopicsSubscriptionsForAllTenantsBrokers() {
        for (TenantInfo tenantInfo : tenantInfoFactory.listTenantInfo()) {
            log.debug("* OqmSubscriberManager on provisioning tenant {}:", tenantInfo.getDataPartitionId());
            //For every Tenant Destination get "subscriberable" Subscriptions
            for (OqmSubscription subscription : getSubscriberableSubscriptions(tenantInfo)) {
                log.debug("* * OqmSubscriberManager on provisioning for tenant {}, subscription {}:", tenantInfo.getDataPartitionId(), subscription.getName());

                //Register a Subscriber on every subscription
                registerSubscriber(tenantInfo, subscription);

                log.debug("* * OqmSubscriberManager on provisioning for tenant {}, subscription {}: Subscriber REGISTERED.", tenantInfo.getDataPartitionId(), subscription.getName());
            }
            log.debug("* OqmSubscriberManager on provisioning tenant {}: COMPLETED.", tenantInfo.getDataPartitionId());
        }
    }

    void provisionControlTopicsWithSubscriptionsAndSubscribersForAllTenantsBrokers() {
        String controlTopicName = properties.getRegisterSubscriberControlTopicName();
        for (TenantInfo tenant : tenantInfoFactory.listTenantInfo()) {

            String tenantId = tenant.getDataPartitionId();

            log.debug("* OqmSubscriberManager on check '{}' subscriber control topic existence at tenant's '{}' message broker.", controlTopicName, tenantId);
            OqmTopic subscriberControlTopic = driver.getTopic(controlTopicName, getDestination(tenant)).orElse(null);
            boolean controlTopicForTenantJustCreated;
            if (subscriberControlTopic != null) {
                log.debug("* * OqmSubscriberManager: '{}' subscriber control topic exists at tenant's '{}' message broker.", controlTopicName, tenantId);
                controlTopicForTenantJustCreated = false;
            } else {
                log.debug("* * OqmSubscriberManager: '{}' subscriber control topic doesn't exist at tenant's '{}' message broker. Trying to create it:", controlTopicName, tenantId);
                driver.createAndGetTopic(controlTopicName, getDestination(tenant));
                controlTopicForTenantJustCreated = true;
            }

            log.debug("* * OqmSubscriberManager on registering '{}' subscriber control topic subscription with name '{}' at tenant's '{}' message broker.",
                    controlTopicName, subscriberControlTopicSubscriptionName, tenantId);

            OqmSubscription subscriberControlTopicSubscriptionForTenant = null;
            OqmSubscription subscriptionRequest = OqmSubscription.builder().name(subscriberControlTopicSubscriptionName)
                    .topics(Collections.singletonList(subscriberControlTopic)).build();

            if (!controlTopicForTenantJustCreated) {
                subscriberControlTopicSubscriptionForTenant = driver.getSubscription(subscriberControlTopicSubscriptionName, getDestination(tenant)).orElse(null);
            }

            if (subscriberControlTopicSubscriptionForTenant != null) {
                log.debug("* * * OqmSubscriberManager on registering '{}' subscriber control topic subscription with name '{}' at tenant's '{}' message broker: Subscription CREATED.",
                        controlTopicName, subscriberControlTopicSubscriptionName, tenantId);
            } else {
                subscriberControlTopicSubscriptionForTenant = driver.createAndGetSubscription(subscriptionRequest, getDestination(tenant));
                log.debug("* * * OqmSubscriberManager on registering '{}' subscriber control topic subscription with name '{}' at tenant's '{}' message broker: Subscription already EXISTS.",
                        controlTopicName, subscriberControlTopicSubscriptionName, tenantId);

                registerControlTopicSubscriber(tenant, subscriberControlTopicSubscriptionForTenant);
                log.debug("* * * OqmSubscriberManager on registering '{}' subscriber control topic subscription with name '{}' at tenant's '{}' message broker: Subscriber REGISTERED.",
                        controlTopicName, subscriberControlTopicSubscriptionName, tenantId);

            }

        }
    }

    private void registerSubscriber(TenantInfo tenantInfo, OqmSubscription subscription) {
        OqmDestination destination = getDestination(tenantInfo);

        OqmMessageReceiver receiver = (oqmMessage, oqmAckReplier) -> {

            String pubsubMessage = oqmMessage.getData();
            String notificationId = subscription.getName();
            Map<String, String> headerAttributes = oqmMessage.getAttributes();


            HttpResponse response;
            boolean ackedNacked = false;
            try {
                response = notificationHandler.notifySubscriber(notificationId, pubsubMessage, headerAttributes);

                if (!response.isSuccessCode()) {
                    log.error(NOT_ACKNOWLEDGE + response.getBody());
                } else {
                    log.debug(ACKNOWLEDGE);
                    oqmAckReplier.ack();
                }
                ackedNacked = true;

            } catch (Exception e) {
                log.debug(NOT_ACKNOWLEDGE, e);
            }

            if (!ackedNacked) oqmAckReplier.nack();
        };

        OqmSubscriber subscriber = OqmSubscriber.builder().subscription(subscription).messageReceiver(receiver).build();
        driver.subscribe(subscriber, destination);
        log.debug("Just subscribed at topic {} subscription {} for tenant {}.",
                subscription.getTopics().get(0), subscription.getName(), tenantInfo.getDataPartitionId());
    }

    private void registerControlTopicSubscriber(TenantInfo tenantInfo, OqmSubscription controlTopicSubscription) {

        OqmMessageReceiver receiver = (oqmMessage, oqmAckReplier) -> {

            String pubsubMessage = oqmMessage.getData();
            Map<String, String> headerAttributes = oqmMessage.getAttributes();
            String subscriptionId = headerAttributes.get("subscription-id");
            String dataPartitionId = headerAttributes.get("data-partition-id");
            String topic = headerAttributes.get("topic");

            OqmSubscription subscription = OqmSubscription.builder()
                    .topic(OqmTopic.builder().name(topic).build())
                    .name(subscriptionId)
                    .build();

            registerSubscriber(tenantInfoFactory.getTenantInfo(dataPartitionId), subscription);

            log.debug(ACKNOWLEDGE);
            oqmAckReplier.ack();
        };

        OqmSubscriber subscriber = OqmSubscriber.builder().subscription(controlTopicSubscription).messageReceiver(receiver).build();
        OqmDestination destination = getDestination(tenantInfo);
        driver.subscribe(subscriber, destination);

        log.debug("Just subscribed at topic {} subscription {} for tenant {}.",
                controlTopicSubscription.getTopics().get(0), controlTopicSubscription.getName(), tenantInfo.getDataPartitionId());
    }

    public List<OqmSubscription> getSubscriberableSubscriptions(TenantInfo tenantInfo) {
        return driver.listSubscriptions(null, INTERESTED_SUBSCRIPTIONS_QUERY, getDestination(tenantInfo));
    }

    private OqmDestination getDestination(TenantInfo tenantInfo) {
        return OqmDestination.builder().partitionId(tenantInfo.getDataPartitionId()).build();
    }

}
