package org.opengroup.osdu.notification.provider.gcp.pubsub;

import org.jetbrains.annotations.NotNull;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentMatchers;
import org.mockito.Mock;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriver;
import org.opengroup.osdu.core.gcp.oqm.model.OqmDestination;
import org.opengroup.osdu.core.gcp.oqm.model.OqmSubscription;
import org.opengroup.osdu.core.gcp.oqm.model.OqmTopic;
import org.opengroup.osdu.notification.provider.gcp.config.OqmConfigurationProperties;
import org.opengroup.osdu.notification.provider.gcp.pubsub.di.OqmNotificationHandler;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;
import org.powermock.reflect.Whitebox;

import java.util.*;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.powermock.api.mockito.PowerMockito.*;

@RunWith(PowerMockRunner.class)
@PrepareForTest(OqmSubscriberManager.class)
public class OqmSubscriberManagerTest {

    public static final String TENANT1 = "tenant1";
    public static final String TENANT2 = "tenant2";
    public static final String TENANT3 = "tenant3";

    @Mock
    ITenantFactory tenantInfoFactory;
    @Mock
    OqmDriver driver;

    @Mock
    OqmNotificationHandler notificationHandler;

    OqmConfigurationProperties properties = new OqmConfigurationProperties();

    OqmSubscriberManager manager;

    private final Long constructDate = System.currentTimeMillis();
    private final String subscriberControlTopicSubscriptionName = "notification-service-" + constructDate;

    @Before
    public void before() throws Exception {
        Collection<TenantInfo> tenants = getMockedTenantInfos();
        when(tenantInfoFactory, "listTenantInfo").thenReturn(tenants);

    }

    @Test
    public void whenProvisionSubscribersOnAllInterestedTopicsSubscriptionsForAllTenantsBrokers_thenProperOperationsInvoked() throws Exception {
        manager = spy(new OqmSubscriberManager(tenantInfoFactory, driver, notificationHandler, properties));

        //Mock 4 subscriberable subscriptions somehow distributed between mock tenants
        doAnswer(question -> {
            TenantInfo tenant = question.getArgument(0);
            String tenantId = tenant.getDataPartitionId();
            switch (tenantId) {
                case TENANT1:
                    return Collections.singletonList(OqmSubscription.builder().name("de-" + tenantId + "-subscription-A").build());
                case TENANT2:
                    return Arrays.asList(
                            OqmSubscription.builder().name("de-" + tenantId + "-subscription-A").build(),
                            OqmSubscription.builder().name("de-" + tenantId + "-subscription-B").build()
                    );
                case TENANT3:
                    return Collections.singletonList(OqmSubscription.builder().name("de-" + tenantId + "-subscription-A").build());
                default:
                    return Collections.emptyList();
            }
        }).when(manager).getSubscriberableSubscriptions(any(TenantInfo.class));

        PowerMockito.doNothing().when(manager, "registerSubscriber", any(TenantInfo.class), any(OqmSubscription.class));

        //Run being tested method.
        manager.provisionSubscribersOnAllInterestedTopicsSubscriptionsForAllTenantsBrokers();

        //verify 3 times getSubscriberableSubscriptions method was called
        verifyPrivate(manager, times(3)).invoke("getSubscriberableSubscriptions", any(TenantInfo.class));

        //verify only 4 subscribers creation were requested and completed
        verifyPrivate(manager, times(4)).invoke("registerSubscriber", any(TenantInfo.class), any(OqmSubscription.class));

    }

    @Test
    public void whenProvisionControlTopicsWithSubscriptionsAndSubscribersForAllTenantsBrokers_thenProperOperationsInvoked() throws Exception {

        //Mock only tenant 2 broker has control topic created so far. Others are not yet.
        when(driver.getTopic(ArgumentMatchers.eq(properties.getRegisterSubscriberControlTopicName()), any(OqmDestination.class)))
                .thenAnswer((question) -> {
                    String topicName = question.getArgument(0);
                    OqmDestination destination = question.getArgument(1);
                    if (destination.getPartitionId().equals(TENANT2)) {
                        return Optional.of(OqmTopic.builder().name(topicName).build());
                    }

                    return Optional.empty();
                });

        //Mock only tenant 2 broker has control topic subscription created so far. Others are not yet.
        when(driver.getSubscription(any(String.class), any(OqmDestination.class)))
                .thenAnswer((question) -> {
                    String subscriptionName = question.getArgument(0);
                    OqmDestination destination = question.getArgument(1);
                    if (destination.getPartitionId().equals(TENANT2)) {
                        return Optional.of(OqmSubscription.builder().name(subscriptionName).build());
                    }

                    return Optional.empty();
                });

        //Mock driver smoothly creates a subscription
        when(driver.createAndGetSubscription(any(OqmSubscription.class), any(OqmDestination.class)))
                .thenAnswer(question -> question.<OqmSubscription>getArgument(0));

        manager = spy(new OqmSubscriberManager(tenantInfoFactory, driver, notificationHandler, properties));
        //Mock subscriberControlTopicSubscriptionName
        Whitebox.setInternalState(manager, "subscriberControlTopicSubscriptionName", subscriberControlTopicSubscriptionName);
        //Mock subscriber silent and successful creation
        PowerMockito.doNothing().when(manager, "registerControlTopicSubscriber", any(TenantInfo.class), any(OqmSubscription.class));

        //Run being tested method.
        manager.provisionControlTopicsWithSubscriptionsAndSubscribersForAllTenantsBrokers();

        //verify 3 times the driver was asked if the topic exists (for all tenants)
        verify(driver, times(3)).getTopic(any(String.class), any(OqmDestination.class));

        //verify 2 times the driver was asked to create and get topic (for tenants 1 and 3
        verify(driver, times(2)).createAndGetTopic(any(String.class), any(OqmDestination.class));

        //verify only 1 time the driver was asked if the subscription exists (for TENANT2)
        verify(driver, times(1)).getSubscription(any(String.class), any(OqmDestination.class));

        //verify 2 times the driver was asked to create the subscription (for TENANT1, TENANT2)
        verify(driver, times(1)).getSubscription(any(String.class), any(OqmDestination.class));

        //verify only 2 subscribers creation were requested and completed
        verifyPrivate(manager, times(2)).invoke("registerControlTopicSubscriber", any(TenantInfo.class), any(OqmSubscription.class));
    }

    @Test
    public void whenUnprovisionControlTopicsSubscriptionsFromAllTenantsBrokers_thenProperOperationsInvoked() {
        //Mock all 3 tenants' brokers have control topic subscriptions created so far.
        when(driver.getSubscription(any(String.class), any(OqmDestination.class)))
                .thenAnswer((question) -> {
                    String subscriptionName = question.getArgument(0);
                    OqmDestination destination = question.getArgument(1);
                    return Optional.of(OqmSubscription.builder().name(subscriptionName).build());
                });

        //Mock driver smoothly deletes a subscription
        PowerMockito.doNothing().when(driver).deleteSubscription(any(String.class), any(OqmDestination.class));

        manager = spy(new OqmSubscriberManager(tenantInfoFactory, driver, notificationHandler, properties));
        //Mock subscriberControlTopicSubscriptionName
        Whitebox.setInternalState(manager, "subscriberControlTopicSubscriptionName", subscriberControlTopicSubscriptionName);

        //Run being tested method.
        manager.unprovisionControlTopicsSubscriptionsFromAllTenantsBrokers();

        //verify 3 times the driver was asked about a subscription existence (for all tenants)
        verify(driver, times(3))
                .getSubscription(ArgumentMatchers.eq(subscriberControlTopicSubscriptionName), any(OqmDestination.class));


        //verify 3 times the driver was asked to delete the subscription (for all tenants)
        verify(driver, times(3))
                .deleteSubscription(ArgumentMatchers.eq(subscriberControlTopicSubscriptionName), any(OqmDestination.class));

    }

    @NotNull
    private Collection<TenantInfo> getMockedTenantInfos() {
        Collection<TenantInfo> tenants = new ArrayList<>();

        //Mock we have 3 tenants
        tenants.add(new TenantInfo() {
            {
                setId(1L);
                setDataPartitionId(TENANT1);
            }
        });
        tenants.add(new TenantInfo() {
            {
                setId(2L);
                setDataPartitionId(TENANT2);
            }
        });
        tenants.add(new TenantInfo() {
            {
                setId(3L);
                setDataPartitionId(TENANT3);
            }
        });
        return tenants;
    }
}